from __future__ import annotations

import argparse
import sys
from typing import Any, Optional

import lazy_import
import rich

from evaluation_system import __version__
from evaluation_system.misc import logger

from .utils import BaseCompleter, BaseParser, standard_main

freva = lazy_import.lazy_module("freva")


class Cli(BaseParser):
    """Class that constructs the Databrowser Argument Parser."""

    desc = "Find data in the system."

    def __init__(
        self,
        parser: Optional[argparse.ArgumentParser] = None,
    ):
        """Construct the databrowser sub arg. parser."""
        super().__init__(parser, "freva-databrowser")
        self.parser.add_argument(
            "--multiversion",
            default=False,
            action="store_true",
            help="Select not only the latest version.",
        )
        self.parser.add_argument(
            "--batch-size",
            default=5000,
            type=int,
            help="Number of files to retrieve.",
        )
        self.parser.add_argument(
            "--count",
            default=False,
            action="store_true",
            help="Show the number of files for each search result.",
        )
        self.parser.add_argument(
            "--facet",
            default=None,
            type=str,
            action="append",
            help=("Retrieve values of given facet instead of files."),
        )
        self.parser.add_argument(
            "--facet-limit",
            type=int,
            help="Limit the number of output facets.",
            default=sys.maxsize,
        )
        self.parser.add_argument(
            "--time-select",
            type=str,
            help="Operator that specifies how the time period is selected.",
            choices=["flexible", "strict", "file"],
            default="flexible",
        )
        self.parser.add_argument(
            "--debug",
            "-v",
            "-d",
            "--verbose",
            help="Use verbose output.",
            action="store_true",
            default=False,
        )
        self.parser.add_argument(
            "facets",
            nargs="*",
            help="Search facet(s).",
            type=str,
            metavar="facets",
        )
        self.parser.set_defaults(apply_func=self.run_cmd)

    @staticmethod
    def run_cmd(
        args: argparse.Namespace,
        **kwargs: Optional[Any],
    ) -> None:
        """Call the databrowser command and print the results."""
        facets: dict[str, Any] = BaseCompleter.arg_to_dict(args.facets, append=True)
        facet_limit = kwargs.pop("facet_limit")
        for key in (
            "facets",
            "facet",
            "count",
            "relevant_only",
            "batch_size",
        ):
            _ = kwargs.pop(key, "")
        for key, values in facets.items():
            if len(values) == 1:
                facets[key] = values[0]
        merged_args: dict[str, Any] = {**kwargs, **facets}
        if args.count:
            out = freva.count_values(facet=args.facet, **merged_args)
        elif args.facet:
            out = freva.facet_search(facet=args.facet, **merged_args)
        else:
            out = freva.databrowser(batch_size=args.batch_size, **merged_args)
        # flush stderr in case we have something pending
        sys.stderr.flush()
        if isinstance(out, dict):
            # We have facet values as return values
            for att, values in out.items():
                facet_limit = facet_limit or len(values) + 1
                try:
                    keys = ",".join(
                        [
                            f"{k} ({c})"
                            for n, (k, c) in enumerate(values.items())
                            if n < facet_limit
                        ]
                    )
                except AttributeError:
                    keys = ",".join(
                        [v for n, v in enumerate(values) if n < facet_limit]
                    )
                if facet_limit < len(values):
                    keys += ",..."
                print(f"{att}: {keys}", flush=True)
            return
        if args.count:
            print(out, flush=True)
        else:
            for key in out:
                print(str(key), flush=True)


def main(argv: Optional[list[str]] = None) -> None:
    standard_main(Cli, __version__, argv)



"""Definitions of custom exceptions and warnings."""

import logging
import sys
import warnings
from contextlib import contextmanager
from typing import Any, Callable

from evaluation_system.misc import logger


def deprecated_method(klass: str, new_method: str):
    """Show a deprecation warning for a deprecated method."""

    def call_deprecated_method(function: Callable[[Any], Any]) -> Callable[[Any], Any]:
        def inner(*args: Any, **kwargs: Any) -> Any:
            if logger.level <= logging.DEBUG:
                raise AttributeError(
                    f"{klass} as no attribute: {function.__name__}, use "
                    f"{new_method} instead"
                )
            msg = (
                f"The `{function.__name__}` of {klass} is deprecated and "
                "might be subject to removal in the future. Please consider "
                f"renaming the `{klass}.{function.__name__}` method to "
                f"`{klass}.{new_method}`"
            )
            logger.warning(msg)
            return function(*args, **kwargs)

        return inner

    return call_deprecated_method


class ConfigurationException(Exception):
    """Mark exceptions thrown in this package"""


class PluginNotFoundError(Exception):
    """Exception definition for missing Plugins."""


class ValidationError(Exception):
    """Thrown if some variable contains an improper value."""


class ParameterNotFoundError(Exception):
    """Thrown if some parameter is not in the database."""


class PluginManagerException(Exception):
    """For all problems generating while using the plugin manager."""


@contextmanager
def hide_exception():
    """
    Suppress traceback, only print the exception type and value
    """
    default_value = getattr(sys, "tracebacklimit", 1000)
    sys.tracebacklimit = 0
    yield
    sys.tracebacklimit = default_value  # revert changes



from __future__ import annotations

import argparse
import sys
from pathlib import Path
from typing import Any, Optional, Type

import lazy_import
import rich

from evaluation_system import __version__
from evaluation_system.misc import logger

from .utils import BaseParser, SubCommandParser, standard_main

UserData = lazy_import.lazy_class("freva.UserData")
freva = lazy_import.lazy_module("freva")

from evaluation_system.misc.exceptions import ValidationError


class IndexData(BaseParser):
    """CLI class that deals with indexing the data."""

    desc = "Index existing user project data to the databrowser."

    def __init__(self, subparser: argparse.ArgumentParser):
        super().__init__(subparser)
        self.parser.add_argument(
            "crawl_dir",
            nargs="*",
            type=Path,
            metavar="crawl_dir",
            help="The user directory(s) that needs to be crawled.",
        )
        self.parser.add_argument(
            "--data-type",
            "--dtype",
            default="fs",
            choices=["fs"],
            help="The data type of the data.",
        )
        self.parser.add_argument(
            "--continue-on-errors",
            "--continue",
            "-c",
            action="store_true",
            help="Continue indexing on error.",
        )
        self.parser.add_argument(
            "--debug",
            "-v",
            "-d",
            "--verbose",
            help="Use verbose output.",
            action="store_true",
            default=False,
        )
        self.parser.set_defaults(apply_func=self.run_cmd)

    @staticmethod
    def run_cmd(args: argparse.Namespace, **kwargs: Any) -> None:
        """Call the crawl my data command and print the results."""
        user_data = UserData()
        try:
            user_data.index(
                *args.crawl_dir,
                dtype=args.data_type,
                continue_on_errors=args.continue_on_errors,
            )
        except (ValidationError, ValueError) as e:
            if args.debug:
                raise e
            try:
                msg = f"{e.__module__}: " f"{e.__str__()}"
            except AttributeError:
                msg = f"{e.__repr__()}"
            logger.error(msg)
            raise SystemExit from e


class AddData(BaseParser):
    """CLI class that deals with indexing the data."""

    desc = "Add new user project data to the databrowser."

    def __init__(self, subparser: argparse.ArgumentParser):
        super().__init__(subparser)
        self.parser.add_argument(
            "product",
            type=str,
            help="Product search key the newly added data can be found.",
        )
        self.parser.add_argument(
            "paths",
            nargs="+",
            type=Path,
            metavar="paths",
            help=(
                "Filename(s) or Directories that are going to be added to the"
                "databrowser."
            ),
        )
        self.parser.add_argument(
            "--how",
            default="copy",
            choices=["copy", "move", "symlink", "link"],
            help=(
                "Method of how the data is added into the central freva user "
                "directory."
            ),
        )
        self.parser.add_argument(
            "--override",
            "--overwrite",
            action="store_true",
            help="Replace existing files in the user data structure.",
            default=False,
        )
        self.parser.add_argument(
            "--project", type=str, default=None, help=argparse.SUPPRESS
        )
        self.parser.add_argument(
            "--experiment",
            type=str,
            default=None,
            help=(
                "Set the <experiment> information if they can't be found in the "
                "meta data."
            ),
        )
        self.parser.add_argument(
            "--institute",
            type=str,
            default=None,
            help=(
                "Set the <institute> information if they can't be found in the "
                "meta data."
            ),
        )
        self.parser.add_argument(
            "--model",
            type=str,
            default=None,
            help=(
                "Set the <model> information if they can't be found in the "
                "meta data."
            ),
        )
        self.parser.add_argument(
            "--variable",
            type=str,
            default=None,
            help=(
                "Set the <variable> information if they can't be found in the "
                "meta data."
            ),
        )
        self.parser.add_argument(
            "--time-frequency",
            "--time_frequency",
            type=str,
            default=None,
            help=(
                "Set the <time_frequency> information if they can't be found in the "
                "meta data."
            ),
        )
        self.parser.add_argument(
            "--ensemble",
            type=str,
            default=None,
            help=(
                "Set the <ensemble> information if they can't be found in the "
                "meta data."
            ),
        )
        self.parser.add_argument(
            "--cmor-table",
            "--cmor_table",
            type=str,
            default=None,
            help=(
                "Set the <cmor-table> information if they can't be found in the "
                "meta data."
            ),
        )
        self.parser.add_argument(
            "--realm",
            type=str,
            default=None,
            help=(
                "Set the <realm> information if they can't be found in the metadata."
            ),
        )
        self.parser.add_argument(
            "--debug",
            "-v",
            "-d",
            "--verbose",
            help="Use verbose output.",
            action="store_true",
            default=False,
        )
        self.parser.set_defaults(apply_func=self.run_cmd)

    @staticmethod
    def run_cmd(args: argparse.Namespace, **kwargs: Any) -> None:
        """Call the crawl my data command and print the results."""
        facets = (
            "experiment",
            "institute",
            "model",
            "variable",
            "time_frequency",
            "ensemble",
            "realm",
        )
        defaults = {k: getattr(args, k) for k in facets if getattr(args, k)}
        defaults["_project"] = kwargs.pop("project", None)
        user_data = UserData()
        try:
            user_data.add(
                args.product,
                *args.paths,
                how=args.how,
                override=args.override,
                **defaults,
            )
        except (ValidationError, ValueError) as e:
            if args.debug:
                raise e
            logger.error("%s", e)
            sys.exit(1)


class DeleteData(BaseParser):
    """CLI class that deals with indexing the data."""

    desc = "Delete existing user project data from the databrowser."

    def __init__(self, subparser: argparse.ArgumentParser):
        self.parser = subparser
        self.parser.add_argument(
            "paths",
            nargs="+",
            type=Path,
            metavar="paths",
            help="The user directory(s) that needs to be crawled.",
        )
        self.parser.add_argument(
            "--delete-from-fs",
            "--delete_from_fs",
            "--delete",
            action="store_true",
            default=False,
            help=(
                "Do not only delete the files from the databrowser but also "
                "from file system."
            ),
        )
        self.parser.add_argument(
            "--debug",
            "-v",
            "-d",
            "--verbose",
            help="Use verbose output.",
            action="store_true",
            default=False,
        )
        self.parser.set_defaults(apply_func=self.run_cmd)

    @staticmethod
    def run_cmd(args: argparse.Namespace, **kwargs: Any) -> None:
        """Call the crawl my data command and print the results."""
        user_data = UserData()
        user_data.delete(*args.paths, delete_from_fs=args.delete_from_fs)


class Cli(SubCommandParser):
    """Class that constructs the Data Crawler Argument Parser."""

    desc = "Update users project data."

    def __init__(
        self,
        parser: Optional[argparse.ArgumentParser] = None,
    ):
        """Construct the esgf sub arg. parser."""
        subcommands: dict[str, Type[BaseParser]] = {
            "index": IndexData,
            "add": AddData,
            "delete": DeleteData,
        }
        super().__init__(parser, sub_parsers=subcommands, command="freva-user-data")
        self.parser.set_defaults(apply_func=self._usage)

    @staticmethod
    def run_cmd(args: argparse.Namespace, **kwargs: str):
        args.apply_func(args, **kwargs)


def main(argv: Optional[list[str]] = None) -> None:
    standard_main(Cli, __version__, argv)



"""
Created on 18.05.2016

@author: Sebastian Illing
"""

import json
import multiprocessing as mp
import os
import time
from functools import partial
from pathlib import Path
from subprocess import Popen

import mock
import pytest

from evaluation_system.misc.exceptions import PluginNotFoundError, ValidationError
from evaluation_system.tests import run_cli, similar_string


def test_cli(dummy_plugin, capsys, dummy_config, caplog):
    with mock.patch("os.getpid", lambda: 12345):
        from evaluation_system.misc import config
        from freva.cli.plugin import main as plugin_cli

        with pytest.raises(SystemExit):
            plugin_cli(["dummyplugin"])
        plugin_cli(["dummyplugin", "the_number=13"])
        output = capsys.readouterr().out
        assert "the_number" in output
        assert "13" in output
        assert os.getpid() == 12345
        out_path = (
            Path(dummy_config.get("scheduler_output_dir"))
            / "dummyplugin"
            / "DummyPlugin-12345.local"
        )
        try:
            out_path.unlink()
        except FileNotFoundError:
            pass
        plugin_cli(["dummyplugin", "the_number=13"])
        interactive = capsys.readouterr().out
        assert dummy_plugin.plugin_output_file == out_path
        assert out_path.exists()
        with out_path.open() as f:
            assert interactive == f.read()
        out_path.unlink()
        plugin_cli(["dummyplugin", "the_number=13", "--batchmode"])
        output = capsys.readouterr().out
        assert "tail -f" in output
        out_f = Path([o.split()[-1] for o in output.split("\n") if "tail" in o][0])
        assert out_f.exists()

        import freva

        plugin_cli(["dummypluginfolders", "variable=pr"])
        folder_path = capsys.readouterr().out.strip().split()[-1]
        assert str(freva.history()[0]["id"]) in os.path.basename(
            os.path.normpath(folder_path)
        )

        plugin_cli(["dummypluginfolders", "variable=pr", "--unique-output", "true"])
        folder_path = capsys.readouterr().out.strip().split()[-1]
        assert str(freva.history()[0]["id"]) in os.path.basename(
            os.path.normpath(folder_path)
        )

        plugin_cli(["dummypluginfolders", "variable=pr", "--unique-output", "false"])
        folder_path = capsys.readouterr().out.strip().split()[-1]
        assert str(freva.history()[0]["id"]) not in os.path.basename(
            os.path.normpath(folder_path)
        )


def test_killed_jobs_set_to_broken(dummy_config):
    import freva

    proc = mp.Process(
        target=freva.run_plugin,
        args=("dummyplugin",),
        kwargs={"the_number": 10, "other": -10},
    )
    proc.start()
    time.sleep(1)
    proc.terminate()
    time.sleep(1)
    hist = freva.history()[0]
    assert hist["status_dict"][hist["status"]].lower() == "broken"


def test_list_tools(capsys, dummy_env):
    from freva.cli.plugin import main as run

    run(["--list-tools"])
    plugin_list = capsys.readouterr().out
    assert "DummyPlugin" in plugin_list
    run_cli(["plugin", "dummyplugin", "--doc"])
    help_str = capsys.readouterr().out
    assert "number" in help_str
    assert "the_number" in help_str
    assert "Option" in help_str
    assert "Description" in help_str


def test_run_pyclientplugin(dummy_history):
    import freva
    from evaluation_system.misc import config

    res = freva.run_plugin("dummyplugin", the_number=32, caption="Some caption")
    assert res.status == "finished"
    assert len(res.version) == 3
    res = freva.plugin_info("dummyplugin", "config", the_number=32)
    res = "\n".join([l.strip() for l in res.split("\n") if l.strip()])
    assert "the_number" in res
    assert "input" in res
    repo = freva.plugin_info("dummyplugin", "repository", the_number=32)
    assert "repository" in repo.lower()
    assert "version" in repo.lower()
    with pytest.raises(ValueError):
        freva.plugin_info("dummyplugin", "repo")

    freva.logger.is_cli = False
    with pytest.raises(PluginNotFoundError):
        freva.run_plugin("dummyplugin0")


def test_plugin_status(dummy_env, caplog) -> None:
    """Test the plugin status queries."""
    import os

    import freva

    res = freva.run_plugin("dummyplugin", the_number=2, other=-5, batchmode=True)
    with pytest.raises(ValueError):
        res.wait(2)
    assert res.status == "running"
    res.kill()
    time.sleep(0.5)
    assert res.status == "broken"
    res.kill()
    res = freva.run_plugin("dummyplugin", the_number=2, other=-1, batchmode=True)
    res.wait()
    assert res.status == "finished"
    assert isinstance(res.batch_id, int)


def test_plugin_output(dummy_history) -> None:
    """Test the output of the plugin."""
    import freva

    res = freva.run_plugin("dummypluginfolders", variable="pr", caption="Some caption")
    assert res.status == "finished"
    assert isinstance(res.get_result_paths(), list)
    assert len(res.get_result_paths()) > 0
    assert "variable" in res.configuration
    assert res.configuration["variable"] == "pr"
    assert len(res.stdout) > 1
    assert res.plugin.lower() == "dummypluginfolders"
    assert len(res.version) == 3
    assert "pr" in res.__repr__()
    assert "finished" in res.__repr__()
    assert "dummypluginfolders" in res.__repr__()
    assert res.job_script == ""
    assert res.batch_id is None


def test_empty_status(dummy_history, capsys) -> None:
    """Test the plugin status queries."""
    import freva

    res = freva.PluginStatus(12345)
    assert res.status == "unknown"
    assert res.configuration == {}
    assert res.stdout == ""
    assert res.batch_id is None
    assert res.job_script == ""
    res.kill()
    assert res.plugin == ""
    assert res.version == (0, 0, 0)
    assert res.get_result_paths() == []


def test_run_plugin(capsys, dummy_history, dummy_env):
    import freva
    from evaluation_system.misc import config

    with pytest.raises(SystemExit):
        run_cli(["plugin", "dummyplugin"])
    # test run tool
    run_cli(
        [
            "plugin",
            "dummyplugin",
            "the_number=32",
            '--caption="Some caption"',
            "--debug",
        ]
    )
    output_str = capsys.readouterr().out
    assert "Dummy tool was run" in output_str
    assert "the_number" in output_str
    # test get version
    run_cli(["plugin", "dummyplugin", "--repo-version"])
    output_str = capsys.readouterr().out
    # test batch mode
    run_cli(["plugin", "dummypluginfolders", "--json"])
    output = json.loads(capsys.readouterr().out)
    assert isinstance(output, dict)
    assert "result" in output
    # test save config
    run_cli(["plugin", "dummyplugin", "the_number=32", "--save", "--debug"])
    fn = (
        Path(config.get(config.BASE_DIR_LOCATION))
        / "config/dummyplugin/dummyplugin.conf"
    )
    assert not fn.is_file()
    # test show config
    run_cli(["plugin", "dummyplugin", "the_number=42", "--show-config"])
    output_str = capsys.readouterr().out
    for line in (
        "number",
        "the_number",
        "something",
        "other",
        "input",
    ):
        assert line in output_str



"""Module to access the esgf data catalogue."""

from __future__ import annotations

from pathlib import Path
from typing import Optional, Union, overload

import lazy_import
from typing_extensions import Literal

from .utils import handled_exception

P2P = lazy_import.lazy_class("evaluation_system.model.esgf.P2P")


__all__ = [
    "esgf_browser",
    "esgf_facets",
    "esgf_datasets",
    "esgf_download",
    "esgf_query",
]


@handled_exception
def esgf_browser(
    opendap: bool = False,
    gridftp: bool = False,
    **search_constraints: dict[str, str],
) -> list[str]:
    """Find data in the ESGF.

    The method queries the ESGF nodes for file URLs (default) or
    opendap/gridftp endpoints as well.

    The ``key=value`` syntax follows that of ``freva.databrowser``
    but the key names follow the ESGF standards for each dataset.
    Search of multiple values for the same key can be achieved
    either as a ``str`` concatenation (e.g.``frequency="3hr,mon" ``
    with **no** space between variables) or as a ``list``
    (e.g. ``frequency=["3hr", "mon"]``).

    Parameters:
    -----------
    opendap: bool, default: False
        List opendap endpoints instead of http ones.
    gridftp: bool, default: False
        Show gridftp endpoints instead of the http default
        ones (or skip them if none found)
    **search_constraints: Union[str, Path, in, list[str]]
        Search facets to be applied in the data search.

    Returns:
    --------
    list :
        Collection of files

    Example
    -------

    Similarly to ``freva.databrowser``, ``freva.esgf_browser``
    expects a list of ``key=value`` pairs in no particular
    order, but unlike the former it *is* case sensitive.

    Given that your Freva instance is configured at DKRZ,
    if we want to search the URLs of all the files stored
    at the (DKRZ) local node (``distrib=false``) holding
    the latest version (``latest=true``) of the variable
    uas (``variable=uas``) for either 3hr or monthly
    time frequencies ( ``frequency=["3hr", "mon"]``)
    and for a particular realization within the
    project ``CMIP6``:

    .. execute_code::

        import freva
        files = freva.esgf_browser(
            mip_era="CMIP6",
            activity_id="ScenarioMIP",
            source_id="CNRM-CM6-1",
            institution_id="CNRM-CERFACS",
            experiment_id="ssp585",
            frequency="3hr,mon",
            variable="uas",
            variant_label="r1i1p1f2",
            distrib=False,
            latest=True,
        )
        print(f"{len(files) =}")
        for file in files[:5]:
            print(file)

    In order to list the opendap endpoints (``opendap=True``):

    .. execute_code::

        import freva
        opendap = freva.esgf_browser(
            mip_era="CMIP6",
            activity_id="ScenarioMIP",
            source_id="CNRM-CM6-1",
            institution_id="CNRM-CERFACS",
            experiment_id="ssp585",
            frequency="3hr",
            variable="uas",
            variant_label="r1i1p1f2",
            distrib=False,
            latest=True,
            opendap=True,
        )
        print(opendap)

    Or the gridftp endpoints instead (``gridftp=True``):

    .. execute_code::
        :hide_code:

        import freva
        gridftp = freva.esgf_browser(
            mip_era="CMIP6",
            activity_id="ScenarioMIP",
            source_id="CNRM-CM6-1",
            institution_id="CNRM-CERFACS",
            experiment_id="ssp585",
            frequency="3hr",
            variable="uas",
            variant_label="r1i1p1f2",
            distrib=False,
            latest=True,
            gridftp=True,
        )
        print(gridftp)
    """
    result_url = []
    url_type = "http"
    if opendap:
        url_type = "opendap"
    if gridftp:
        url_type = "gridftp"
    p2p = P2P()
    for result in p2p.files(fields="url", **search_constraints):
        for url_encoded in result["url"]:
            url, _, access_type = url_encoded.split("|")
            if access_type.lower().startswith(url_type):
                result_url.append(url)
    return result_url


@handled_exception
def esgf_facets(
    show_facet: Optional[Union[str, list[str]]] = None,
    **search_constraints: dict[str, str],
) -> dict[str, list[str]]:
    """Search for data attributes (facets) through ESGF.

    The method queries the ESGF nodes for available search facets (keys)
    like model, experiment etc. The ``key=value`` syntax follows that of
    ``freva.facet_search`` but the key names follow the ESGF standards
    for each dataset.

    Parameters:
    -----------
    show_facet: Union[str, list[str]], default: None
        List all values for the given facet (might be
        defined multiple times). The results show the possible
        values of the selected facet according to the given
        constraints and the number of *datasets* (not files)
        that selecting such value as a constraint will result
        (faceted search)
    **search_constraints: Union[str, Path, in, list[str]]
        Search facets to be applied in the data search.

    Returns:
    --------
    dict[str, list[str]]:
        Collection of facets

    Example
    -------

    List the values of the attributes ``variable``
    and ``time_frequency`` stored at the (DKRZ) local node
    (``distrib=false``) holding the latest version (``latest=true``)
    for a particular realization within the project ``CMIP6``:

    .. execute_code::

        import freva
        facets = freva.esgf_facets(
            mip_era="CMIP6",
            activity_id="ScenarioMIP",
            source_id="CNRM-CM6-1",
            institution_id="CNRM-CERFACS",
            experiment_id="ssp585",
            distrib=False,
            latest=True,
            show_facet=["variable", "frequency"])
        print(facets)

    """
    show_facet = show_facet or []
    p2p = P2P()
    return p2p.get_facets(show_facet, **search_constraints)


@handled_exception
def esgf_datasets(
    **search_constraints: dict[str, str],
) -> list[tuple[str, str]]:
    """List the name of the datasets (and version) in the ESGF.

    The method queries the ESGF nodes for dataset information.
    The ``key=value`` syntax follows that of ``freva.databrowser``
    but the key names follow the ESGF standards for each dataset.

    Parameters:
    -----------
    **search_constraints: Union[str, Path, in, list[str]]
        Search facets to be applied in the data search.

    Returns:
    --------
    list[tuple[str, str]]:
        list of (dataset_name, version_number) tuples

    Example
    -------

    List the datasets corresponding to a query of files
    stored at the (DKRZ) local node (``distrib=false``)
    holding the latest version (``latest=true``) for a particular
    realization within the project ``CMIP6``:

    .. execute_code::

        import freva
        datasets = freva.esgf_datasets(
            mip_era="CMIP6",
            activity_id="ScenarioMIP",
            source_id="CNRM-CM6-1",
            institution_id="CNRM-CERFACS",
            experiment_id="ssp585",
            frequency="3hr",
            variable="uas",
            variant_label="r1i1p1f2",
            distrib=False,
            latest=True,
            )
        print(f"{len(datasets) =}")
        for dataset in datasets[:5]:
            print(dataset)
    """
    p2p = P2P()
    return sorted(p2p.get_datasets_names(**search_constraints))


@handled_exception
def esgf_download(
    download_script: Optional[Union[str, Path]] = None,
    **search_constraints: dict[str, str],
) -> Union[str, Path]:
    """Create a script file to download the queried files at ESGF.

    The method creates a bash script wrapper of a wget query
    from ESGF dataset(s) (only http).

    Parameters:
    -----------
    download_script: Union[str, Path], default: None
        Download wget_script for getting the files
        instead of displaying anything (only http)
    **search_constraints: Union[str, Path, in, list[str]]
        Search facets to be applied in the data search.

    Returns:
    --------
    Union[str, Path]:
        wget_script to download the files (only http).

    Example
    -------

    Create a wget script to download the queried URLs:

    .. execute_code::

        import freva
        freva.esgf_download(project="CMIP5",
                    experiment="decadal1960",
                    variable="tas", distrib=False, latest=True,
                    download_script="/tmp/script.get")

        with open('/tmp/script.get', 'r') as f:
            content = f.readlines()
            print(" ".join(content[:10]))

    .. note::

        You will need an OpenID account to download the data,
        for example `here <https://esgf-data.dkrz.de/user/add/?next=http://esgf-data.dkrz.de/projects/esgf-dkrz/>`_.

        There is a `ESGF PyClient <https://esgf-pyclient.readthedocs.io/en/latest/index.html>`_ as well.


    """
    p2p = P2P()
    if not download_script:
        import random
        import string

        random_string = "".join(
            random.choices(string.ascii_letters + string.digits, k=16)
        )
        base_path = "/tmp/script"
        extension = "get"
        download_script = f"{base_path}-{random_string}.{extension}"
    download_script = Path(download_script)
    download_script.touch(0o755)
    with download_script.open("bw") as f:
        f.write(p2p.get_wget(**search_constraints))
    return str(f"Download script successfully saved to {download_script}")


@handled_exception
def esgf_query(
    query: Optional[Union[str, list[str]]] = None,
    **search_constraints: dict[str, str],
) -> list[dict[str, list[str]]]:
    """Query fields from ESGF and group them per dataset

    The method queries fields (e.g. facets) and groups them by dataset
    in a list of dictionaries.

    Parameters:
    -----------
    query:
        Display results from <list> queried fields
    **search_constraints:
        Search facets to be applied in the data search.

    Returns:
    --------
    list[dict[str, list[str]]]:
        List of dictionaries with the queried elements for each dataset

    Example:
    --------

    Show the ``key=value`` pair of selected elements for a search (e.g.
    ``query=["url","master_id","distribution","mip_era","activity_id","source_id","variable","product","version"]``):

    .. execute_code::

        import json, freva
        queries = freva.esgf_query(
            mip_era="CMIP6",
            activity_id="ScenarioMIP",
            source_id="CNRM-CM6-1",
            institution_id="CNRM-CERFACS",
            experiment_id="ssp585",
            frequency="3hr",
            variant_label="r1i1p1f2",
            distrib=False,
            latest=True,
            query=["url","master_id","distribution",
                "mip_era","activity_id","source_id",
                "variable","product","version"],
        )
        print(json.dumps(queries[:2], indent=3))
    """
    p2p = P2P()
    if isinstance(query, list):
        query = ",".join(query)
    return p2p.get_datasets(fields=query, **search_constraints)



"""Submit jobs using the moab workload manager."""

from __future__ import annotations

from typing import ClassVar

from .pbs import PBSJob


class MoabJob(PBSJob):
    submit_command: ClassVar[str] = "msub"
    cancel_command: ClassVar[str] = "canceljob"
    config_name: ClassVar[str] = "moab"



from copy import deepcopy
from pathlib import Path
from typing import Any, Dict, Optional

import django
from django.conf import settings
from evaluation_system.misc import config
from evaluation_system.misc.config import reloadConfiguration

SETTINGS = dict(
    TIME_ZONE="UTC",
    USE_TZ=False,
    ATOMIC_REQUESTS=True,
    CONN_HEALTH_CHECKS=True,
)
# DB_LOADED: bool = False
# DB_OVERRIDDEN: bool = False
config.reloadConfiguration()
# Application definition
SETTINGS["INSTALLED_APPS"] = (
    "django.contrib.auth",  # We need this to access user groups
    "django.contrib.flatpages",
    "django.contrib.contenttypes",
    "django.contrib.sites",
)
SETTINGS["DATABASES"] = {
    "default": {
        "ENGINE": "django.db.backends.mysql",
        "NAME": config.get(config.DB_DB) or "",
        "USER": config.get(config.DB_USER) or "",
        "PASSWORD": config.get(config.DB_PASSWD) or "",
        "HOST": config.get(config.DB_HOST)
        or "",  # Or an IP Address that your DB is hosted on
        "PORT": config.get(str(config.DB_PORT), "3306"),
    }
}

try:
    settings.configure(**SETTINGS)
    django.setup()
except RuntimeError:
    pass


def reconfigure_django(config_file: Optional[Path] = None) -> None:
    """Reconfigure the django settings."""
    django_settings: Dict[str, Any] = deepcopy(SETTINGS)
    reloadConfiguration(config_file)
    django_settings["DATABASES"] = {
        "default": {
            "ENGINE": "django.db.backends.mysql",
            "NAME": config.get(config.DB_DB) or "",
            "USER": config.get(config.DB_USER) or "",
            "PASSWORD": config.get(config.DB_PASSWD) or "",
            "HOST": config.get(config.DB_HOST) or "",
            "PORT": config.get(str(config.DB_PORT), "3306"),
        }
    }
    try:
        settings.configure(**django_settings)
        django.setup()
    except RuntimeError:
        for key, value in django_settings["DATABASES"]["default"].items():
            settings.DATABASES["default"][key] = value



from __future__ import annotations

import argparse
import json
import sys
from contextlib import contextmanager
from pathlib import Path
from typing import Any, Iterator, Optional

from rich import print as pprint
from rich.console import Console

console = Console()
import lazy_import

from evaluation_system import __version__
from evaluation_system.misc import logger

from .utils import BaseCompleter, BaseParser, standard_main

freva = lazy_import.lazy_module("freva")
PluginNotFoundError = lazy_import.lazy_class(
    "evaluation_system.misc.exceptions.PluginNotFoundError"
)
ParameterNotFoundError = lazy_import.lazy_class(
    "evaluation_system.misc.exceptions.ParameterNotFoundError"
)
ValidationError = lazy_import.lazy_class(
    "evaluation_system.misc.exceptions.ValidationError"
)
hide_exception = lazy_import.lazy_function(
    "evaluation_system.misc.exceptions.hide_exception"
)


@contextmanager
def _pipe(redirect_stdout: bool = False) -> Iterator[None]:
    """Redirect the stdout to stderr if needed."""

    stdout = sys.stdout
    try:
        if redirect_stdout is True:
            sys.stdout = sys.stderr
        yield
    finally:
        sys.stdout = stdout


class Cli(BaseParser):
    """Class that constructs the Plugin Argument Parser."""

    desc = "Apply data analysis plugin."

    def __init__(
        self,
        parser: Optional[argparse.ArgumentParser] = None,
    ):
        """Construct the plugin sub arg. parser."""
        super().__init__(parser, "freva-plugin")
        self.parser.add_argument(
            "tool-name",
            nargs="?",
            metavar="plugin_name",
            help="Plugin name.",
            default=None,
        )
        self.parser.add_argument(
            "--repo-version",
            default=False,
            action="store_true",
            help="Show the version number from the repository.",
        )
        self.parser.add_argument(
            "--caption",
            default="",
            help="Set a caption for the results",
        )
        self.parser.add_argument(
            "--save",
            default=False,
            action="store_true",
            help="Save the plugin configuration to default destination.",
        )
        self.parser.add_argument(
            "--save-config",
            type=Path,
            default=None,
            help="Save the plugin configuration.",
        )
        self.parser.add_argument(
            "--show-config",
            help="Show the resulting configuration.",
            action="store_true",
            default=False,
        )
        self.parser.add_argument(
            "--scheduled-id",
            default=None,
            type=int,
            help=argparse.SUPPRESS,
        )
        self.parser.add_argument(
            "--batchmode",
            help="Create a Batch job and submit it to the scheduling system.",
            default=False,
            action="store_true",
        )
        self.parser.add_argument(
            "--unique-output",
            "--unique_output",
            choices=["true", "false", "True", "False"],
            help="Append a Freva run id to the output/cache folder(s).",
            default="true",
        )
        self.parser.add_argument(
            "--debug",
            "-v",
            "-d",
            "--verbose",
            help="Use verbose output.",
            action="store_true",
            default=False,
        )
        self.parser.add_argument(
            "--list-tools",
            "--list",
            "-l",
            default=False,
            action="store_true",
            help="Only list the available tools.",
        )
        self.parser.add_argument(
            "--doc",
            "--plugin-doc",
            default=False,
            action="store_true",
            help="Display plugin documentation",
        )
        self.parser.add_argument(
            "--json",
            "-j",
            help=(
                "Display a json representation of the result, this can be"
                "useful if you want to build shell based pipelines and want"
                "parse the output with help of `jq`."
            ),
            default=False,
            action="store_true",
        )
        self.parser.add_argument(
            "--wait",
            "-w",
            help=(
                "Wait for the plugin to finish, this has only an effect for "
                "batch mode execution."
            ),
            default=False,
            action="store_true",
        )

        self.parser.add_argument("tool-options", nargs="*", help="Tool options")
        self.parser.set_defaults(apply_func=self.run_cmd)

    @staticmethod
    def run_cmd(
        args: argparse.Namespace,
        **kwargs: Any,
    ) -> None:
        """Call the plugin command and print the results."""
        tool_name = kwargs.pop("tool-name")
        jsonify = kwargs.pop("json", False)
        wait = kwargs.pop("wait", False)
        tool_options = kwargs.pop("tool-options", []) + kwargs.pop("unknown", [])
        repo_version: bool = kwargs.pop("repo_version", False)
        show_config: bool = kwargs.pop("show_config", False)
        if kwargs.pop("list_tools") or not tool_name:
            console.print(freva.get_tools_list())
            return
        options: dict[str, Any] = BaseCompleter.arg_to_dict(tool_options)
        for key, val in options.items():
            if len(val) == 1:
                options[key] = val[0]
        if kwargs["unique_output"].lower() == "true":
            kwargs["unique_output"] = True
        else:
            kwargs["unique_output"] = False
        tool_args = {**kwargs, **options}
        value = 0
        try:
            if tool_args.pop("doc"):
                console.print(freva.plugin_doc(tool_name))
                return
            if repo_version:
                print(freva.plugin_info(tool_name or "", "repository", **options))
            elif show_config:
                print(freva.plugin_info(tool_name or "", "config", **options))
            else:
                with _pipe(redirect_stdout=jsonify):
                    plugin_run = freva.run_plugin(tool_name or "", **tool_args)
                value = int(plugin_run.status == "broken")
        except (
            PluginNotFoundError,
            ValidationError,
            ParameterNotFoundError,
        ) as e:
            if args.debug:
                raise e
            with hide_exception():
                logger.error(e)
                raise SystemExit
        if value != 0:
            logger.warning("Tool failed to run")
        elif jsonify or wait:
            plugin_run.wait()
        if jsonify:
            print(plugin_run)


def main(argv: Optional[list[str]] = None) -> None:
    standard_main(Cli, __version__, argv)



"""Test loading different configurations."""

from pathlib import Path
from tempfile import TemporaryDirectory

import pytest

import freva

from .conftest import get_config
from .mocks import DUMMY_PATH


def test_load_config(dummy_config) -> None:
    """Test loading a different config."""
    cfg, _ = get_config()
    with TemporaryDirectory() as temp_dir:
        eval_config = Path(temp_dir) / "evaluation_system.conf"
        cfg.set("evaluation_system", "solr.host", "foo.bar.com.au")
        with open(eval_config, "w") as f_obj:
            cfg.write(f_obj)
        assert dummy_config.get("solr.host") != "foo.bar.com.au"
        with freva.config(eval_config):
            assert dummy_config.get("solr.host") == "foo.bar.com.au"
        assert dummy_config.get("solr.host") != "foo.bar.com.au"
        with freva.config(eval_config):
            with pytest.raises(ValueError):
                freva.run_plugin("dummyplugin", the_number=1, batchmode=True)
            with pytest.warns(UserWarning):
                freva.run_plugin("dummyplugin", the_number=1)


def test_load_user_plugin(dummy_config) -> None:
    """Test loading a user defined plugin."""
    with freva.config(plugin_path=f"{DUMMY_PATH},dummy"):
        assert "dummyplugin" in freva.list_plugins()
    assert "dummyplugin" not in freva.list_plugins()



import pytest

from evaluation_system.api.workload_manager.core import Job


def test_get_cancel_command():
    from evaluation_system.api.workload_manager import cancel_command

    with pytest.raises(NotImplementedError):
        cancel_command("bla", 1000)
    assert "scancel" in cancel_command("slurm", 1000)


def test_parse_bytes():

    from evaluation_system.api.workload_manager.core import parse_bytes

    assert parse_bytes(123) == 123, "Integer input should return the input as is"

    assert (
        parse_bytes(123.0) == 123
    ), "Float input should return the input converted to int"

    assert (
        parse_bytes("100") == 100
    ), "String input with digits should be correctly parsed"

    assert (
        parse_bytes(" 100 MB ") == 100000000
    ), "String input with spaces should be correctly parsed"

    assert parse_bytes("5kB") == 5000, "String with kB should be correctly parsed"
    assert parse_bytes("1kiB") == 1024, "String with kiB should be correctly parsed"

    assert (
        parse_bytes("MB") == 1000000
    ), "String without digits should prepend '1' and parse correctly"

    try:
        parse_bytes("5 blah")
        assert False, "Expected ValueError for input with invalid units"
    except ValueError as e:
        assert (
            str(e) == "Could not interpret 'blah' as a byte unit"
        ), "Expected specific ValueError message for invalid units"

    try:
        parse_bytes("1e6 blah")
        assert False, "Expected ValueError for malformed number input"
    except ValueError as e:
        assert "Could not interpret" in str(
            e
        ), "Expected ValueError for malformed number input"


def test_string_to_bytes():
    from evaluation_system.api.workload_manager.core import string_to_bytes

    assert (
        string_to_bytes("5g") == 5 * 1000**3
    )  # Specific case when byte_form length is 1


def test_get_format_bytes():
    from evaluation_system.api.workload_manager.core import format_bytes

    assert format_bytes(1) == "1 B", "Failed to format bytes correctly."
    assert format_bytes(1023) == "1.00 kiB", "Failed to format Kibibytes correctly."
    assert format_bytes(1024) == "1.00 kiB", "Failed to format Kibibytes correctly."
    assert format_bytes(2**20) == "1.00 MiB", "Failed to format Mebibytes correctly."
    assert format_bytes(2**30) == "1.00 GiB", "Failed to format Gibibytes correctly."
    assert format_bytes(2**40) == "1.00 TiB", "Failed to format Tebibytes correctly."
    assert format_bytes(2**50) == "1.00 PiB", "Failed to format Pebibytes correctly."
    assert format_bytes(1024 * 1024 - 1) == "1.00 MiB", "Failed on edge of MiB."


class MockJob(Job):
    submit_command = "mock_submit"
    cancel_command = "mock_cancel"

    def __init__(self, *args, **kwargs):
        # For simplicity, bypass Job's __init__ with a direct call to super().__init__()
        # This avoids needing to deal with Job's abstract __init__ parameters.
        super(Job, self).__init__(*args, **kwargs)
        self.job_id = ""  # Ensuring this attribute is set for testing.
        self.name = "mock_job"


# def test_job_methods():

#     job_instance = MockJob()

#     with patch("subprocess.Popen") as mock_popen:
#         mock_process = MagicMock()
#         mock_process.communicate.return_value = (b"output", b"")
#         mock_process.returncode = 0
#         mock_popen.return_value = mock_process

#         output = job_instance._call(["echo", "test"])
#         assert output == "output", "Expected mocked output from _call method."

#     job_id_output = "Submitted batch job 12345"
#     assert (
#         job_instance._job_id_from_submit_output(job_id_output) == "12345"
#     ), "Failed to extract job ID."

#     try:
#         job_instance._job_id_from_submit_output("No job id here")
#     except ValueError as e:
#         assert "Could not parse job id" in str(
#             e
#         ), "Expected failure when job ID is missing."

#     with patch.object(MockJob, "_call", return_value="") as mock_call:
#         job_instance = MockJob()
#         job_instance._close_job("12345", "scancel")
#         mock_call.assert_called_with(["scancel", "12345"])



from __future__ import annotations

import argparse
import json
import sys
from pathlib import Path
from typing import Any, Optional, Union, cast

import lazy_import
import rich

from evaluation_system import __version__

from .utils import BaseParser, standard_main

freva = lazy_import.lazy_module("freva")
BaseCompleter = lazy_import.lazy_class("freva.cli.utils.BaseCompleter")


class Cli(BaseParser):
    """Class that constructs the ESGF Query Argument Parser."""

    desc = "Search/Download ESGF the data catalogue."

    def __init__(
        self,
        parser: Optional[argparse.ArgumentParser] = None,
    ):
        """Construct the esgf sub arg. parser."""
        super().__init__(parser, "freva-esgf")
        self.parser = parser or argparse.ArgumentParser(
            prog="freva-esgf",
            description=self.desc,
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        )
        self.parser.add_argument(
            "--datasets",
            default=False,
            action="store_true",
            help="List the name of the datasets instead of showing the urls.",
        )
        self.parser.add_argument(
            "--show-facet",
            default=None,
            action="append",
            help=(
                "List all values for the given facet (might be "
                "defined multiple times). The results show the possible "
                "values of the selected facet according to the given "
                "constraints and the number of *datasets* (not files) "
                "that selecting such value as a constraint will result "
                "(faceted search)."
            ),
        )
        self.parser.add_argument(
            "--opendap",
            default=False,
            action="store_true",
            help="Show opendap endpoints instead of http ones.",
        )
        self.parser.add_argument(
            "--gridftp",
            default=False,
            action="store_true",
            help=(
                "Show gridftp endpoints instead of the http default "
                "ones (or skip them if none found)."
            ),
        )
        self.parser.add_argument(
            "--download-script",
            default=None,
            type=Path,
            help=(
                "Download wget_script for getting the files "
                "instead of displaying anything (only http)."
            ),
        )
        self.parser.add_argument(
            "--query",
            default=None,
            type=str,
            help=("Query fields from ESGF and group them per dataset."),
        )
        self.parser.add_argument(
            "--debug",
            "-v",
            "-d",
            "--verbose",
            help="Use verbose output.",
            action="store_true",
            default=False,
        )
        self.parser.add_argument(
            "facets",
            nargs="*",
            help="Search facet(s).",
            type=str,
            metavar="facets",
        )
        self.parser.set_defaults(apply_func=self.run_cmd)

    @staticmethod
    def run_cmd(
        args: argparse.Namespace,
        other_args: Optional[list[str]] = None,
        **kwargs: Any,
    ) -> None:
        """Call the esgf command and print the results."""
        facets: dict[str, Union[list[str], str]] = {}
        _facets = BaseCompleter.arg_to_dict(args.facets)
        _ = kwargs.pop("facets", None)
        for key, val in _facets.items():
            if len(val) == 1:
                facets[key] = val[0]
            else:
                facets[key] = val
        merged_args = cast(Any, {**kwargs, **facets})
        vars_to_pop = [
            "datasets",
            "download_script",
            "show_facet",
            "opendap",
            "gridftp",
            "query",
        ]
        if args.datasets:
            _ = [merged_args.pop(variable, "") for variable in vars_to_pop]
            out = freva.esgf_datasets(**merged_args)
            print("\n".join([f"{d[0]} - version: {d[1]}" for d in out]))
        elif args.show_facet:
            vars_to_pop.remove("show_facet")
            _ = [merged_args.pop(variable, "") for variable in vars_to_pop]
            out = freva.esgf_facets(**merged_args)
            for facet_key in sorted(out):
                if len(out[facet_key]) == 0:
                    values = "<No Results>"
                else:
                    values = "\n\t".join(
                        [
                            "%s: %s" % (k, out[facet_key][k])
                            for k in sorted(out[facet_key])
                        ]
                    )
                    print("[%s]\n\t%s" % (facet_key, values))
        elif args.download_script:
            vars_to_pop.remove("download_script")
            _ = [merged_args.pop(variable, "") for variable in vars_to_pop]
            out = freva.esgf_download(**merged_args)
            print(out)
        elif args.query:
            vars_to_pop.remove("query")
            _ = [merged_args.pop(variable, "") for variable in vars_to_pop]
            out = freva.esgf_query(**merged_args)
            if len(args.query.split(",")) > 1:
                print(json.dumps(out, indent=3))
            else:
                print("\n".join([str(d) for d in list(out)]))
        else:
            vars_to_pop.remove("opendap")
            vars_to_pop.remove("gridftp")
            _ = [merged_args.pop(variable, "") for variable in vars_to_pop]
            out = freva.esgf_browser(**merged_args)
            print("\n".join([str(d) for d in out]))


def main(argv: Optional[list[str]] = None) -> None:
    standard_main(Cli, __version__, argv)
